searchNodes=[{"doc":"Sibyl is a library which augments the BEAM's default tracing capabilities by hooking into :telemetry , :dbg (the BEAM's built in tracing and debugging functionality), and OpenTelemetry . Basic Usage To leverage all that Sibyl gives you, you need to use it in a module like so: defmodule MyApp.Users do use Sibyl end Following this, you're able to begin emitting telemetry events and tracing function execution. Tracing Function Execution Sibyl provides two decorators which you can use in your modules to automatically trace function execution. These are @decorate_all trace() and @decorate trace() , which automatically traces all functions in the given module, or the function most immediately following the decorator respectively. Sibyl's function tracing follows :telemetry 's standard specification for capturing spans. At the beginning of a function a :start event is emitted; and at the end of a function a :stop event is emitted. If an exception (arising from a raise or throw ) is detected, an :exception event is emitted instead. Unlike :telemetry.span/3 however, Sibyl's trace decorators inline these event emissions into the compiled bytecode of your module, which is slightly more efficient than wrapping traces within anonymous functions. This also has the benefit of making stacktraces easier to read. Event names are automatically determined such that a :start event is emitted by the function MyApp.Users.sign_up ; the [:my_app, :users, :sign_up, :start] event is emitted. Examples follow: defmodule MyApp.Users do use Sibyl @decorate trace ( ) def sign_up ( attrs ) do :ok end end defmodule MyApp.Mailer do use Sibyl @decorate_all trace ( ) def build_mail ( attrs ) do :ok end def send_mail ( attrs ) do attrs |&gt; build_mail ( ) |&gt; handle_send ( ) end end Custom Event Emission Sibyl also provides a thin wrapper over :telemetry.execute/3 to make event emission less error prone. Sibyl will perform compile time checks prior to attempting to emit an event to make sure that it has previously been defined in the current module. Events which are defined in a module are automatically prefixed with that module's namespace such that given a module MyApp.Users and an event :registered , the resultant event will be compiled and ultimately emitted as [:my_app, :users, :registered] . Otherwise, you can specify a module which defined a given event by using Sibyl.emit/4 over Sibyl.emit/3 . For more in-depth information, see Sibyl.emit/4 . Examples follow: defmodule MyApp.Users do use Sibyl def sign_up ( attrs ) do emit :registered # Fails to compile as event is unknown. end end defmodule MyApp.Users do use Sibyl define_event :registered def sign_up ( attrs ) do emit :registered # Compiles properly and emits event end end defmodule MyApp.Users.Mailer do use Sibyl def send_email do emit MyApp.Users , :registered # Compiles properly and emits events emit MyApp.Users , :invalid_event # Causes error as this was not defined in specified module end end Reflection Due to the fact that Sibyl is able to check whether or not events have been defined prior to use, Sibyl exposes a reflection API in the way of Sibyl.Events.reflect/0 and Sibyl.Events.reflect/1 . Please see documentation for Sibyl.Events if you're interested in more potential avenues for extending Sibyl or metaprogramming. Telemetry Handlers Building on top of Sibyl's reflection API, we are able to provide functions to automatically attach defined events to :telemetry handlers. You can do this via the helper functions in Sibyl.Handlers . Please see the documentation for Sibyl.Handlers for more information, but a brief usage example follows: @impl Sibyl.Handler def handle_event ( event , measurement , metadata , opts ) do IO . inspect ( { event , measurement , metadata , opts } ) end :ok = Sibyl.Handlers . attach_all_events ( __MODULE__ ) Additionally, Sibyl provides two example :telemetry handlers: a very basic Elixir Logger handler for demonstration purposes, as well as an OpenTelemetry handler which was what prompted the building of Sibyl in the first place. OpenTelemetry Integration OpenTelemetry is a widely understood specification for dealing with event emission and traces. One can use OpenTelemetry to be able to instrument your code base with events and traces quite easily. However; the Elixir community also very much utilises :telemetry as the standard telemetry/metric/span gathering library. It is possible, of course, to use OpenTelemetry as well as :telemetry , but it would be convinient to have one unified API which bridges both worlds. Very much inspired by the OpentelemetryTelemetry library, Sibyl provides a generic :telemetry handler which bridges any :telemetry.span/3 -spec complaint events to OpenTelemetry traces. Simply attach the Sibyl.Handlers.OpenTelemetry handler and start tracing functions, any captured traces will be handled by your configured OLTP exporter of choice. For demonstration purposes, this project also contains a docker-compose.yml which sets up Jaeger : an easy to use distributed tracing UI to view spans which understands OpenTelemetry . Dynamic Tracing In order to aide debugging of running systems without needing to instrument your code with decorators or event emission, Sibyl also provides an experimental dynamic tracer which leverages the BEAM's built in tracing functionalities. Please make sure you understand what turning on the BEAM's tracing functionality can do to overload a production system however. This is not neccessarily advisable, but it is possible. After enabling Sibyl's dynamic tracer and attaching a :telemetry handler, all future invokations of any given functions will be handled as though you had instrumented your codebase with @decorate trace() . Examples follow: { :ok , _meta } = Sibyl.Dynamic . enable ( Sibyl.Handlers.OpenTelemetry ) { :ok , _meta } = Sibyl.Dynamic . trace ( Enum , :map , 2 ) Enum . map ( [ 1 , 2 , 3 ] , &amp; &amp;1 ) # Produces OpenTelemetry traces :ok = Sibyl.Dynamic . disable ( )","ref":"Sibyl.html","title":"Sibyl","type":"module"},{"doc":"Emits an event. Can be called in one of two modes, either: Sibyl.emit/4 or Sibyl.emit/3 . When called via the arity-4 variant, expects the following arguments to be passed in: A module alias such as Enum or MyApp.Telemetry . An atom: the name of the event you wish to emit, which should be defined in the given module. An optional map containing &quot;measurements&quot; for the given event. An optional map containing &quot;metadata&quot; for the given event. When called via the arity-3 variant, expects the following arguments to be passed in: An event, either in the form of a singular atom, or in the form of a list of atoms. If a singular atom is passed in, it is expected that the given atom is defined as as event in the caller's module. If a list of atoms is passed in, no expectations hold as Sibyl assumes you're trying to emit an event that was not defined by Sibyl . An optional map containing &quot;measurements&quot; for the given event. An optional map containing &quot;metadata&quot; for the given event. No other combination of arguments is supported and an error will be raised at compile time if called any other way than described. When called properly, Sibyl will perform a compile-time check on the event you're attempting to emit (unless the event was specified as a list of atoms), to ensure that the event was properly defined by Sibyl before use.","ref":"Sibyl.html#emit/4","title":"Sibyl.emit/4","type":"macro"},{"doc":"Utility module for working with ASTs","ref":"Sibyl.AST.html","title":"Sibyl.AST","type":"module"},{"doc":"Returns true if the given argument is an Elixir AST node representing a module alias such as Enum .","ref":"Sibyl.AST.html#alias?/1","title":"Sibyl.AST.alias?/1","type":"macro"},{"doc":"Given an alias AST node, returns the fully resolved alias that said node would expand to. For example, given: {:__aliases, unused(), [Elixir, Enum]} , returns: Enum .","ref":"Sibyl.AST.html#module/2","title":"Sibyl.AST.module/2","type":"function"},{"doc":"Returns the :__unused__ atom.","ref":"Sibyl.AST.html#unused/0","title":"Sibyl.AST.unused/0","type":"macro"},{"doc":"Returns true if the given argument is equal to :__unused__ . Primarily used internally.","ref":"Sibyl.AST.html#unused?/1","title":"Sibyl.AST.unused?/1","type":"macro"},{"doc":"","ref":"Sibyl.AST.html#t:alias/0","title":"Sibyl.AST.alias/0","type":"type"},{"doc":"","ref":"Sibyl.AST.html#t:ast/0","title":"Sibyl.AST.ast/0","type":"type"},{"doc":"","ref":"Sibyl.AST.html#t:unused/0","title":"Sibyl.AST.unused/0","type":"type"},{"doc":"Module encapsulating Sibyl's business logic for decorating functions for automated tracing. Should be used only via use Sibyl . When used like this, provides the ability to decorate your function calls with: @decorate trace() to automatically trace a single function's execution, emitting telemetry events for :start , :end , and :exception sub-events. @decorate_all trace() which does the same as the above, but automatically for all functions in a module. Automatically traced functions are available for reflection by via Sibyl.Events.reflect/1 .","ref":"Sibyl.Decorator.html","title":"Sibyl.Decorator","type":"module"},{"doc":"","ref":"Sibyl.Decorator.html#trace/0","title":"Sibyl.Decorator.trace/0","type":"macro"},{"doc":"Decorator which wraps a given function with a standard telemetry span. The name of the captured event will be determined by however Sibyl is configured to generate event names. Due to how anonymous functions are defined and executed in the BEAM, it ends up being quite a bit more performant to build the span manually rather than using :telemetry.span/3 . See here for an example of how to emit the correct events. See here for explanations w.r.t. anonymous function perf.","ref":"Sibyl.Decorator.html#trace/2","title":"Sibyl.Decorator.trace/2","type":"function"},{"doc":"Module which contains functions which allow you to bridge together the built in BEAM debugging and tracer with modules implementing the Sibyl.Handler behaviour. This provides extremely powerful functionality as you can effectively instruct the BEAM to start building OpenTelemetry traces without code instrumentation on production environments. It is important to note that leveraging the BEAM's built in debugging and tracing functionality can have severe memory and CPU requirements so it isn't something one should do lightly. Regardless, in order to track down heisenbugs and other maladies which can only be reproduced on production environments, this functionality should prove extremely useful.","ref":"Sibyl.Dynamic.html","title":"Sibyl.Dynamic","type":"module"},{"doc":"","ref":"Sibyl.Dynamic.html#disable/0","title":"Sibyl.Dynamic.disable/0","type":"function"},{"doc":"","ref":"Sibyl.Dynamic.html#enable/1","title":"Sibyl.Dynamic.enable/1","type":"function"},{"doc":"","ref":"Sibyl.Dynamic.html#trace/3","title":"Sibyl.Dynamic.trace/3","type":"function"},{"doc":"Utils module contianing various custom guards in the scope of Sibyl.Dynamic","ref":"Sibyl.Dynamic.Guards.html","title":"Sibyl.Dynamic.Guards","type":"module"},{"doc":"Returns true if given term is a message sent by the Erlang :dbg module for function and function-return traces.","ref":"Sibyl.Dynamic.Guards.html#trace?/1","title":"Sibyl.Dynamic.Guards.trace?/1","type":"macro"},{"doc":"Returns true if given term is a message sent by the Erlang :dbg module and if the message type corresponds to the given type.","ref":"Sibyl.Dynamic.Guards.html#type?/2","title":"Sibyl.Dynamic.Guards.type?/2","type":"macro"},{"doc":"Module containing the core business logic of Sibyl. Includes utility functions to defining events, emitting events, and reflecting on events which are defined in any module loaded on the BEAM.","ref":"Sibyl.Events.html","title":"Sibyl.Events","type":"module"},{"doc":"Given an atom denoting some abstract event (but not of type event() ), builds an event name such that the given abstract event is formatted as per telemetry best practices. Examples (assuming this is being called in a module and function) iex&gt; Sibyl.Events . build_event ( :not_found ) [ :my_app , :&quot;some_function/2&quot; , :not_found ]","ref":"Sibyl.Events.html#build_event/1","title":"Sibyl.Events.build_event/1","type":"macro"},{"doc":"Builds a consistent telemetry event name following the conventions specified in this post . Called like: Sibyl.build_event(MyApp.Accounts, :register_user, 2, :email_sent) , produces the following event: [:my_app, :accounts, :&quot;register_user/2&quot;, :email_sent]","ref":"Sibyl.Events.html#build_event/4","title":"Sibyl.Events.build_event/4","type":"function"},{"doc":"Defines the given event. Events should only be used once they are defined. Unless this is done, reflect/0 and reflect/1 will fail to see said event and will eventually throw errors. When given a singular atom, it is assumed that the event you are trying to define is namespaced to said module as: [:my_app, :my_module, :custom_event] . When given a list of atoms, the event is simply registered as whatever you passed in; i.e. [:some, :custom, :event] .","ref":"Sibyl.Events.html#define_event/2","title":"Sibyl.Events.define_event/2","type":"macro"},{"doc":"See emit/4","ref":"Sibyl.Events.html#emit/3","title":"Sibyl.Events.emit/3","type":"function"},{"doc":"Emits the given event. Note: this is a low level API which should not be called outside of Sibyl's own code, and largely exists to plumb together the system. Please prefer to use Sibyl.emit/3 which includes compile time checks to make sure the events being emitted were registered. When given an event as an atom, tries to emit the event namespaced to the given module (see define_event/2 for more information). When given a event as a list, simply emits that given list.","ref":"Sibyl.Events.html#emit/4","title":"Sibyl.Events.emit/4","type":"function"},{"doc":"Returns true if event was defined.","ref":"Sibyl.Events.html#is_event/1","title":"Sibyl.Events.is_event/1","type":"function"},{"doc":"Given a module that may, or may not, use Sibyl as well as an event, returns true if said module defines the given event.","ref":"Sibyl.Events.html#is_event/2","title":"Sibyl.Events.is_event/2","type":"function"},{"doc":"Returns a list of all telemetry events which have been defined by Sibyl.Events","ref":"Sibyl.Events.html#reflect/0","title":"Sibyl.Events.reflect/0","type":"function"},{"doc":"Returns a list of all telemetry events which have been defined by Sibyl.Events for the given module. Note: some events may be implicitly defined via the top level @decorate trace() or @decorate_all trace() decorators.","ref":"Sibyl.Events.html#reflect/1","title":"Sibyl.Events.reflect/1","type":"function"},{"doc":"","ref":"Sibyl.Events.html#t:event/0","title":"Sibyl.Events.event/0","type":"type"},{"doc":"","ref":"Sibyl.Events.html#t:sibyl_event/0","title":"Sibyl.Events.sibyl_event/0","type":"type"},{"doc":"Behaviour for :telemetry handlers","ref":"Sibyl.Handler.html","title":"Sibyl.Handler","type":"behaviour"},{"doc":"","ref":"Sibyl.Handler.html#c:handle_event/4","title":"Sibyl.Handler.handle_event/4","type":"callback"},{"doc":"Groups functions to make it easy to attach telemetry events to handlers","ref":"Sibyl.Handlers.html","title":"Sibyl.Handlers","type":"module"},{"doc":"Reflects upon the state of the current application's modules and all other dynamically loaded module and attaches any events defined in those modules to the given handler. Any options are forwarded to the given handler also. Takes an optional, but recommended :name =&gt; String.t() option too.","ref":"Sibyl.Handlers.html#attach_all_events/2","title":"Sibyl.Handlers.attach_all_events/2","type":"function"},{"doc":"Attaches the given events the given handler. Any options are forwarded to the given handler also. Takes an optional, but recommended :name =&gt; String.t() option too.","ref":"Sibyl.Handlers.html#attach_events/3","title":"Sibyl.Handlers.attach_events/3","type":"function"},{"doc":"Reflects upon the given module and attaches any events defined in those modules to the given handler. Any options are forwarded to the given handler also. Takes an optional, but recommended :name =&gt; String.t() option too.","ref":"Sibyl.Handlers.html#attach_module_events/3","title":"Sibyl.Handlers.attach_module_events/3","type":"function"},{"doc":"","ref":"Sibyl.Handlers.html#t:handler/0","title":"Sibyl.Handlers.handler/0","type":"type"},{"doc":"An example Telemetry handler converting :telemetry events into Chrome-compatible flamegraphs. Exposes two additional functions when compared to traditional Sibyl.Handler s: start/0 which instructs Elixir.Sibyl.Handlers.FlameGraph to start persisting metadata in order to build the resultant flamegraph. stop/1 which instructs Elixir.Sibyl.Handlers.FlameGraph to flush any captured metadata into a JSON file for further use. If you're using a Chrome-derived browser, you'll be able to open and introspect generated flamegraphs via the Chrome Tracing builtin. Otherwise, you can use open source tools which understand Chrome Tracing's format to render graphs. Examples of open source apps that can render Chrome Traces include Speedscope and Perfetto . Note that this utility has only been tested and confirmed working with telemetry explicitly passed in via Sibyl and may not work when listening to arbitrary :telemetry events passed in from other OTP applications. Notes on the file format used by Google's tracer, Speedscope, and other related tools can be found here .","ref":"Sibyl.Handlers.FlameGraph.html","title":"Sibyl.Handlers.FlameGraph","type":"module"},{"doc":"Instructs Elixir.Sibyl.Handlers.FlameGraph to start capturing and persisting :telemetry metadata.","ref":"Sibyl.Handlers.FlameGraph.html#start/0","title":"Sibyl.Handlers.FlameGraph.start/0","type":"function"},{"doc":"Returns true if Elixir.Sibyl.Handlers.FlameGraph has been started prior to invokation.","ref":"Sibyl.Handlers.FlameGraph.html#started?/0","title":"Sibyl.Handlers.FlameGraph.started?/0","type":"function"},{"doc":"Instructs Elixir.Sibyl.Handlers.FlameGraph to output any captured :telemetry metadata into the JSON file of your choice.","ref":"Sibyl.Handlers.FlameGraph.html#stop/1","title":"Sibyl.Handlers.FlameGraph.stop/1","type":"function"},{"doc":"An example Telemetry handler for logging given events to the current application's configured logger. Please note that this very much is an example . There is little reason to do this in production code.","ref":"Sibyl.Handlers.Logger.html","title":"Sibyl.Handlers.Logger","type":"module"},{"doc":"OpenTelemetry is an open source standard telemetry standard which allows us to capture custom metrics and traces of our application. For local development, you can install tools such as Jaeger to be able to view and test OpenTelemetry traces. However, much of the BEAM ecosystem (and this library) uses :telemetry as a standard for emitting arbitrary telemetry events. This handler is a bridge between standard :telemetry span events and OpenTelemetry spec compliant traces. Any event which ends in :start will start an OpenTelemetry span context. Any event which ends in :stop or :exception will stop the currently active OpenTelemetry span context, capturing any metadata that is passed in. Any event which ends in anything else will be attached as a custom event to the currently active span context. For Elixir's built in Task module, Sibyl.Handlers.OpenTelemetry will be able to internal state and automatically attach said async Task to the original parent. Distributed traces via non Task means is also supported, but this cannot be automated. To opt into this behaviour, serialize the span state on the consumer end via the build_distributed_trace_context/0 function, and prior to starting any new traces on the consumer end, use attach_distributed_trace_context/1 . Doing this should attach both the consumer and producer side of the trace into one for rendering in tools such as Jaeger.","ref":"Sibyl.Handlers.OpenTelemetry.html","title":"Sibyl.Handlers.OpenTelemetry","type":"module"},{"doc":"Processes a serialized OpenTelemetry trace context (obtained via build_distributed_trace_context/0 ) and persists it within the current process. The next span which is created will be automatically attached to this distributed trace context, allowing for the building of (tooling permitting) distributed traces across multiple different nodes, releases, services, etc.","ref":"Sibyl.Handlers.OpenTelemetry.html#attach_distributed_trace_context/1","title":"Sibyl.Handlers.OpenTelemetry.attach_distributed_trace_context/1","type":"function"},{"doc":"Serializes the current OpenTelemetry trace context to allow it to be sent over-the-wire to external services, with the intention to re-attach the context on the consumer's end. This should allow you (tooling permitting) to build distributed traces across multiple different nodes, releases, services, etc.","ref":"Sibyl.Handlers.OpenTelemetry.html#build_distributed_trace_context/0","title":"Sibyl.Handlers.OpenTelemetry.build_distributed_trace_context/0","type":"function"},{"doc":"A constant for the undefined OpenTelemetry trace context. Useful for defaulting a distributed trace context parse to a &quot;noop&quot; if data is malformed or mistransmitted.","ref":"Sibyl.Handlers.OpenTelemetry.html#undefined_trace_context/0","title":"Sibyl.Handlers.OpenTelemetry.undefined_trace_context/0","type":"function"},{"doc":"","ref":"Sibyl.BadEmissionError.html","title":"Sibyl.BadEmissionError","type":"exception"},{"doc":"","ref":"Sibyl.UndefinedEventError.html","title":"Sibyl.UndefinedEventError","type":"exception"}]